﻿namespace tSNE

open Aardvark.Base
open FsAlg.Generic
open System

open MatrixFun

/// simple tSNE algorithm like the pseudocode in the paper
module Algorithm =

    /// gradient computation 
    let gradient (ps : Matrix<double>) (qs :Matrix<double>) (tSne : Matrix<double>) (ys : Matrix<double>) n =
        let L = (ps - qs) .* tSne // (pij - qij) * (1 + |yi - yj|²)^-1
        let diag = L |> MatrixFun.colSum |> MatrixFun.diag n 
        4.0 * (diag - L) * ys // 4 * Σ * (yi - yj)

    /// simple version of tSNE (pseudecode paper)
    let tSNEsimpleVersion perp T (η : double) (n : int) (sqDis : Matrix<double>) =
        
        let r = Random()
        
        // pariwise affinites pji (probabilities for high dim)
        let mutable highProbs = HighDimProb.Calculate n perp sqDis

        let mutable gDescent = Matrix.init n 2 ( fun _ _ -> 0.0 ) 
        let mutable gain = Matrix.init n 2 ( fun _ _ -> 1.0 )
        let mutable α = 0.5
        
        // start with random normal distributed low dimensional data
        let nextRandomGaussian (r : Random) = 
            let mutable gr = Double.NaN
            while gr.IsNaN() do
                let u = r.NextDouble() * 2.0 - 1.0
                let v = r.NextDouble() * 2.0 - 1.0
                let r2 =  u * u + v * v
                let c = sqrt(-2.0 * log(r2) / r2)
                gr <-  u * c
            gr
            
        // initialize log dim data with random gaussian distributed data
        let mutable ys = Matrix.init n 2 ( fun _ _ -> 
            (r |> nextRandomGaussian) * 0.0001
        )

        Report.BeginTimed("starting tSne computation ...\n")
        for t = 1 to T-1 do
            
            // iteration output
            if t % 100 = 0 then printfn "gradient iteration #%i" t
                
            // compute low dimensional affinities
            let (tDis, probs) = LowDimProb.Calculate n ys

            // gradient
            let g = gradient highProbs tDis probs ys n

            // update information gain
            gain <- 
                let s1 = MatrixFun.sign(g)
                let s2 = MatrixFun.sign(gDescent)
                let e =  MatrixFun.equal s1 s2 
                let ne = MatrixFun.notEqual s1 s2 
                (gain + 0.2) .* ne + (gain * 0.8) .* e
                
            // computation like paper pseudo code
            gDescent <- α * gDescent - η * (gain .* g)
            ys <- ys + gDescent
            
            // mean normalizing again
            let mean = ys |> MatrixFun.meanCol |> Matrix.toArray |> MatrixFun.ofArray n
            ys <- ys - mean

            // increase momentum term
            if t = 250 then 
                α <- 0.8

            // stop lying about high probabilities (optimization)
            if t = 100 then
                highProbs <- highProbs / 4.0
            
        Report.End() |> ignore
        ys