﻿namespace tSNE

open Aardvark.Base
open FsAlg.Generic
open System

open MatrixFun

/// computation of high-dimensional propability matrix
module HighDimProb =

    /// computation of high-dimensional propability matrix
    let Calculate n tPerp (distances : Matrix<double>) = 
        
        let mutable Probs = Matrix.init 0 n (fun _ _ -> 0.0) // probability matrix
        let logPerp = log(tPerp) 

        // computing high dim probabilities and entropy
        let hAndProb (D : Vector<double>) (s : double) = 
            
            // exp( - distance(x_i, x_j) * sigma )
            let prob = -D * s |> Vector.toArray |> Array.map( fun x -> exp(x) )
            let sumProb = prob |> Array.sum

            let e = (D .* (prob |> vector)) |> Vector.toArray |> Array.sum
            let H = log(sumProb) + s * e / sumProb // entropy: H = -∑ p_i * log(p_i)
            
            let P = (prob |> vector) / sumProb // equation 1: calculation of high dim probabilities
            (H,P)
        
        for i = 0 to n-1 do

            // appropriate σ is found by a binary search
            // small σ for dense regions
            let mutable σ = 1.0
            let mutable σ_min = Double.NegativeInfinity
            let mutable σ_max = Double.PositiveInfinity
            
            // filter 0.0 disances (distance from point to itself)
            let d = distances.[i,*] |> Matrix.toArray |> Array.filter ( fun x -> x <> 0.0) |> vector
            let (H, _) = hAndProb d σ
            let mutable Hdiff = H - logPerp // difference of entropy (for stopping condition)

            let mutable j = 0
            let mutable p = Array.empty<double> |> vector
            while abs(Hdiff) > Double.Epsilon && j < 50 do
                
                j <- j + 1
                
                // binary search of sigma
                σ <-  
                    match Hdiff > 0.0 with
                    | true -> 
                        σ_min <- σ
                        match σ_max.IsInfinity() with
                        | true ->  σ_min * 2.0 
                        | false -> (σ_min + σ_max) / 2.0
                    | false -> 
                        σ_max <- σ
                        match σ_min.IsInfinity() with
                        | true -> σ_max / 2.0
                        | false -> (σ_max + σ_min) / 2.0
               
                // update Hdiff for next iteration
                let (H, currProp) = hAndProb d σ
                Hdiff <- H - logPerp
               
                // insert diagonal 0.0 values again (distance from point to itself) 
                p <- Array.init (currProp.Length + 1) (fun k ->
                        match k = i with
                        | true -> 0.0
                        | false -> 
                            match k < i with
                            | true -> currProp.[k]
                            | false -> currProp.[k-1]
                    ) |> vector
                
            Probs <- Probs |> Matrix.appendRow p

        // probabilities have to sum to 1 --> so force it!
        let ps = 
            Probs 
            |> Matrix.toRows
            |> Seq.map (fun r -> r / (Probs |> MatrixFun.sum) |> Matrix.toSeq)
            |> matrix
            
        // create symmetric matrix: (pij + pji) / 2 
        // adding factor 4 --> paper: optimization!
        Matrix.init n n (fun i j -> (0.5 * (ps.[i,j] + ps.[j,i])) * 4.0 )

