﻿namespace MatrixFun

open Aardvark.Base
open FsAlg.Generic

/// some matrix operations 
module MatrixFun =

    /// normalize input data like in the original tSNE algorithm
    let normalize (m : Matrix<double>) =
        let m1 = m - (m |> Matrix.toArray |> Array.min)
        m1 / (m1 |> Matrix.toArray |> Array.max)
   
    /// returns sum of given matrix
    let sum (m : Matrix<double>) =
        m |> Matrix.toSeq |> Seq.sum
    
    /// returns sum of each row
    let rowSum (m : Matrix<double>) =
        m |> Matrix.toRows |> Seq.map (fun i -> i |> sum) |> Seq.toArray

    /// returns sum of each column
    let colSum (m : Matrix<double>) =
        m |> Matrix.toCols |> Seq.map (fun i -> i |> sum) |> Seq.toArray

    /// returns sign matrix
    let sign (m : Matrix<double>) = 
        m |> Matrix.map ( fun i -> 
            match i > 0.0 with
            | true -> 1.0
            | false -> 
                match (i.IsTiny()) with
                | true -> 0.0
                | false -> -1.0
        ) 

    /// returns size of given matrix
    let size (m : Matrix<double>) =
        (m.Rows,m.Cols)
    
    /// returns matrix with element wise equality
    let equal (m : Matrix<double>) (n : Matrix<double>) =
        let (x,y) = size m
        Matrix.init x y (fun i j ->
            let d = m.[i,j] - n.[i,j] |> abs
            match (d.IsTiny()) with
            | true -> 1.0
            | false -> 0.0
        )

    /// return matrix with element wise inequality
    let notEqual (m : Matrix<double>) (n : Matrix<double>) =
        let (x,y) = size m
        Matrix.init x y (fun i j ->
            let d = m.[i,j] - n.[i,j] |> abs
            match (d.IsTiny()) with
            | true -> 0.0
            | false -> 1.0
        )
    
    /// converts given matrix to array of vectors
    let toVectorArray n (m : Matrix<double>) =
        let a = m.ToArray()
        Array.init n (fun i -> a.[i] |> vector)

    /// returns squared distances matrix 
    let squaredDistances n (m : Matrix<double>) = 
        Matrix.init n n ( fun i j -> 
            let qi = m |> Matrix.row i
            let qj = m |> Matrix.row j
            qi - qj |> Matrix.toArray |> vector |> Vector.l2normSq
        ) 

    /// creates diagonal matrix with given values
    let diag n (d : double[]) =
        Matrix.init n n (fun i j ->
            match i = j with
            | true -> d.[i]
            | false -> 0.0
        )
    
    /// creates a matrix of given array
    let ofArray n (a : double[]) =
        Matrix.init n a.Length (fun _ j -> a.[j])

    /// returns mean of matrix
    let mean (m : Matrix<double>) = 
        (m |> Matrix.toArray).Mean()
    
    /// returns means of each column
    let meanCol (m : Matrix<double>) =
        let n = m |> Matrix.toCols |> Seq.map (fun i -> i |> mean) |> Seq.toArray
        Matrix.init (n.Length) 1 (fun i _ -> n.[i])
    
    /// returns means of each row
    let meanRow (m : Matrix<double>) =
        let n = m |> Matrix.toRows |> Seq.map (fun i -> i |> mean) |> Seq.toArray
        Matrix.init (n.Length) 1 (fun i _ -> n.[i])

    /// returns max value of given matrix
    let maxValue (m : Matrix<double>) =
        m |> Matrix.toArray |> Array.max
        

